@@ -187,6 +187,28 @@ class SignedDigestAlgorithmId(ObjectIdentifier):
         '1.2.840.10045.4': 'ecdsa',
     }
 
+    _reverse_map = {
+        'dsa': '1.2.840.10040.4.1',
+        'ecdsa': '1.2.840.10045.4',
+        'md2_rsa': '1.2.840.113549.1.1.2',
+        'md5_rsa': '1.2.840.113549.1.1.4',
+        'rsassa_pkcs1v15': '1.2.840.113549.1.1.1',
+        'rsassa_pss': '1.2.840.113549.1.1.10',
+        'sha1_dsa': '1.2.840.10040.4.3',
+        'sha1_ecdsa': '1.2.840.10045.4.1',
+        'sha1_rsa': '1.2.840.113549.1.1.5',
+        'sha224_dsa': '2.16.840.1.101.3.4.3.1',
+        'sha224_ecdsa': '1.2.840.10045.4.3.1',
+        'sha224_rsa': '1.2.840.113549.1.1.14',
+        'sha256_dsa': '2.16.840.1.101.3.4.3.2',
+        'sha256_ecdsa': '1.2.840.10045.4.3.2',
+        'sha256_rsa': '1.2.840.113549.1.1.11',
+        'sha384_ecdsa': '1.2.840.10045.4.3.3',
+        'sha384_rsa': '1.2.840.113549.1.1.12',
+        'sha512_ecdsa': '1.2.840.10045.4.3.4',
+        'sha512_rsa': '1.2.840.113549.1.1.13',
+    }
+
 
 class SignedDigestAlgorithm(Sequence):
     _fields = [
