@@ -16,7 +16,6 @@ Other type classes are defined that help compose the types listed above.
 from __future__ import unicode_literals, division, absolute_import, print_function
 
 from encodings import idna  # noqa
-import codecs
 import hashlib
 import re
 import socket
@@ -25,8 +24,9 @@ import sys
 import unicodedata
 
 from ._errors import unwrap
+from ._iri import iri_to_uri, uri_to_iri
 from ._ordereddict import OrderedDict
-from ._types import type_name, str_cls, byte_cls
+from ._types import type_name, str_cls
 from .algos import SignedDigestAlgorithm
 from .core import (
     Any,
@@ -60,22 +60,9 @@ from .keys import PublicKeyInfo
 from .util import int_to_bytes, int_from_bytes, inet_ntop, inet_pton
 
 if sys.version_info < (3,):
-    from urlparse import urlsplit, urlunsplit
-    from urllib import (
-        quote as urlquote,
-        unquote as unquote_to_bytes,
-    )
-
     bytes_to_list = lambda byte_string: [ord(b) for b in byte_string]
 
 else:
-    from urllib.parse import (
-        quote as urlquote,
-        unquote_to_bytes,
-        urlsplit,
-        urlunsplit,
-    )
-
     bytes_to_list = list
 
 
@@ -87,6 +74,7 @@ else:
 class DNSName(IA5String):
 
     _encoding = 'idna'
+    _bad_tag = 19
 
     def __ne__(self, other):
         return not self == other
@@ -129,99 +117,11 @@ class URI(IA5String):
 
         self._normalized = True
         self._native = value
-        self.contents = self._normalize(value)
+        self.contents = iri_to_uri(value)
         self._header = None
         if self._trailer != b'':
             self._trailer = b''
 
-    def _normalize(self, value):
-        """
-        Normalizes and encodes a unicode IRI into an ASCII byte string
-
-        :param value:
-
-        """
-
-        parsed = urlsplit(value)
-
-        scheme = _urlquote(parsed.scheme)
-        hostname = parsed.hostname
-        if hostname is not None:
-            hostname = hostname.encode('idna')
-        username = _urlquote(parsed.username)
-        password = _urlquote(parsed.password)
-        port = parsed.port
-        if port is not None:
-            port = str_cls(port).encode('ascii')
-
-        netloc = b''
-        if username is not None:
-            netloc += username
-            if password:
-                netloc += b':' + password
-            netloc += b'@'
-        if hostname is not None:
-            netloc += hostname
-        if port is not None:
-            default_http = scheme == b'http' and port == b'80'
-            default_https = scheme == b'https' and port == b'443'
-            if not default_http and not default_https:
-                netloc += b':' + port
-
-        path = _urlquote(parsed.path, safe='/')
-        query = _urlquote(parsed.query, safe='&=')
-        fragment = _urlquote(parsed.fragment)
-
-        if query is None and fragment is None and path == b'/':
-            path = None
-
-        # Python 2.7 compat
-        if path is None:
-            path = ''
-
-        output = urlunsplit((scheme, netloc, path, query, fragment))
-        if isinstance(output, str_cls):
-            output = output.encode('latin1')
-        return output
-
-    def __unicode__(self):
-        """
-        :return:
-            A unicode string
-        """
-
-        parsed = urlsplit(self.contents)
-
-        scheme = parsed.scheme
-        if scheme is not None:
-            scheme = scheme.decode('ascii')
-
-        username = _urlunquote(parsed.username, remap=[':', '@'])
-        password = _urlunquote(parsed.password, remap=[':', '@'])
-        hostname = parsed.hostname
-        if hostname:
-            hostname = hostname.decode('idna')
-        port = parsed.port
-        if port:
-            port = port.decode('ascii')
-
-        netloc = ''
-        if username is not None:
-            netloc += username
-            if password:
-                netloc += ':' + password
-            netloc += '@'
-        if hostname is not None:
-            netloc += hostname
-        if port is not None:
-            netloc += ':' + port
-
-        path = _urlunquote(parsed.path, remap=['/'], preserve=True)
-        query = _urlunquote(parsed.query, remap=['&', '='], preserve=True)
-        fragment = _urlunquote(parsed.fragment)
-
-        return urlunsplit((scheme, netloc, path, query, fragment))
-
     def __ne__(self, other):
         return not self == other
 
@@ -239,7 +139,15 @@ class URI(IA5String):
         if not isinstance(other, URI):
             return False
 
-        return self._normalize(self.native) == self._normalize(other.native)
+        return iri_to_uri(self.native) == iri_to_uri(other.native)
+
+    def __unicode__(self):
+        """
+        :return:
+            A unicode string
+        """
+
+        return uri_to_iri(self.contents)
 
 
 class EmailAddress(IA5String):
@@ -534,12 +442,14 @@ class NameType(ObjectIdentifier):
         '2.5.4.12': 'title',
         '2.5.4.15': 'business_category',
         '2.5.4.17': 'postal_code',
+        '2.5.4.20': 'telephone_number',
         '2.5.4.41': 'name',
         '2.5.4.42': 'given_name',
         '2.5.4.43': 'initials',
         '2.5.4.44': 'generation_qualifier',
         '2.5.4.46': 'dn_qualifier',
         '2.5.4.65': 'pseudonym',
+        '2.5.4.97': 'organization_identifier',
         # https://tools.ietf.org/html/rfc2985#page-26
         '1.2.840.113549.1.9.1': 'email_address',
         # Page 10 of https://cabforum.org/wp-content/uploads/EV-V1_5_5.pdf
@@ -548,6 +458,8 @@ class NameType(ObjectIdentifier):
         '1.3.6.1.4.1.311.60.2.1.3': 'incorporation_country',
         # https://tools.ietf.org/html/rfc2247#section-4
         '0.9.2342.19200300.100.1.25': 'domain_component',
+        # http://www.alvestrand.no/objectid/0.2.262.1.10.7.20.html
+        '0.2.262.1.10.7.20': 'name_distinguisher',
     }
 
     # This order is largely based on observed order seen in EV certs from
@@ -575,8 +487,11 @@ class NameType(ObjectIdentifier):
         'name',
         'pseudonym',
         'dn_qualifier',
+        'telephone_number',
         'email_address',
         'domain_component',
+        'name_distinguisher',
+        'organization_identifier',
     ]
 
     @property
@@ -599,6 +514,7 @@ class NameType(ObjectIdentifier):
             'title': 'Title',
             'business_category': 'Business Category',
             'postal_code': 'Postal Code',
+            'telephone_number': 'Telephone Number',
             'name': 'Name',
             'given_name': 'Given Name',
             'initials': 'Initials',
@@ -610,7 +526,9 @@ class NameType(ObjectIdentifier):
             'incorporation_state_or_province': 'Incorporation State/Province',
             'incorporation_country': 'Incorporation Country',
             'domain_component': 'Domain Component',
-        }[self.native]
+            'name_distinguisher': 'Name Distinguisher',
+            'organization_identifier': 'Organization Identifier',
+        }.get(self.native, self.native)
 
 
 class NameTypeAndValue(Sequence):
@@ -633,6 +551,7 @@ class NameTypeAndValue(Sequence):
         'title': DirectoryString,
         'business_category': DirectoryString,
         'postal_code': DirectoryString,
+        'telephone_number': PrintableString,
         'name': DirectoryString,
         'given_name': DirectoryString,
         'initials': DirectoryString,
@@ -646,6 +565,8 @@ class NameTypeAndValue(Sequence):
         'incorporation_state_or_province': DirectoryString,
         'incorporation_country': DirectoryString,
         'domain_component': DNSName,
+        'name_distinguisher': DirectoryString,
+        'organization_identifier': DirectoryString,
     }
 
     _prepped = None
@@ -932,7 +853,7 @@ class Name(Choice):
     _sha256 = None
 
     @classmethod
-    def build(cls, name_dict):
+    def build(cls, name_dict, use_printable=False):
         """
         Creates a Name object from a dict of unicode string keys and values.
         The keys should be from NameType._map, or a dotted-integer OID unicode
@@ -942,11 +863,21 @@ class Name(Choice):
             A dict of name information, e.g. {"common_name": "Will Bond",
             "country_name": "US", "organization": "Codex Non Sufficit LC"}
 
+        :param use_printable:
+            A bool - if PrintableString should be used for encoding instead of
+            UTF8String. This is for backwards compatiblity with old software.
+
         :return:
             An x509.Name object
         """
 
         attributes = []
+        if not use_printable:
+            encoding_name = 'utf8_string'
+            encoding_class = UTF8String
+        else:
+            encoding_name = 'printable_string'
+            encoding_class = PrintableString
 
         for attribute_name in NameType.preferred_order:
             if attribute_name not in name_dict:
@@ -963,8 +894,8 @@ class Name(Choice):
                 )
             else:
                 value = DirectoryString(
-                    name='utf8_string',
-                    value=UTF8String(name_dict[attribute_name])
+                    name=encoding_name,
+                    value=encoding_class(name_dict[attribute_name])
                 )
 
             attributes.append(NameTypeAndValue({
@@ -1016,8 +947,10 @@ class Name(Choice):
                 for type_val in rdn:
                     field_name = type_val['type']
                     if field_name in self._native:
-                        self._native[field_name] = [self._native[field_name]]
-                        self._native[field_name].append(type_val['value'])
+                        existing = self._native[field_name]
+                        if not isinstance(existing, list):
+                            existing = self._native[field_name] = [existing]
+                        existing.append(type_val['value'])
                     else:
                         self._native[field_name] = type_val['value']
         return self._native
@@ -1031,31 +964,24 @@ class Name(Choice):
 
         if self._human_friendly is None:
             data = OrderedDict()
+            last_field = None
             for rdn in self.chosen:
                 for type_val in rdn:
                     field_name = type_val['type'].human_friendly
+                    last_field = field_name
                     if field_name in data:
                         data[field_name] = [data[field_name]]
                         data[field_name].append(type_val['value'])
                     else:
                         data[field_name] = type_val['value']
             to_join = []
-            for key in data:
+            keys = data.keys()
+            if last_field == 'Country':
+                keys = reversed(list(keys))
+            for key in keys:
                 value = data[key]
-                if isinstance(value, list):
-                    native_sub_values = []
-                    for sub_value in value:
-                        if isinstance(sub_value, list):
-                            native_sub_value = ', '.join(
-                                reversed([sub_sub_value.native for sub_sub_value in sub_value])
-                            )
-                        else:
-                            native_sub_value = sub_value.native
-                        native_sub_values.append(native_sub_value)
-                    value = ', '.join(reversed(native_sub_values))
-                else:
-                    value = value.native
-                to_join.append('%s: %s' % (key, value))
+                native_value = self._recursive_humanize(value)
+                to_join.append('%s: %s' % (key, native_value))
 
             has_comma = False
             for element in to_join:
@@ -1068,6 +994,23 @@ class Name(Choice):
 
         return self._human_friendly
 
+    def _recursive_humanize(self, value):
+        """
+        Recursively serializes data compiled from the RDNSequence
+
+        :param value:
+            An Asn1Value object, or a list of Asn1Value objects
+
+        :return:
+            A unicode string
+        """
+
+        if isinstance(value, list):
+            return', '.join(
+                reversed([self._recursive_humanize(sub_value) for sub_value in value])
+            )
+        return value.native
+
     @property
     def sha1(self):
         """
@@ -1572,8 +1515,8 @@ class CertificatePolicies(SequenceOf):
 
 class PolicyMapping(Sequence):
     _fields = [
-        ('issuer_domain_policy', ObjectIdentifier),
-        ('subject_domain_policy', ObjectIdentifier),
+        ('issuer_domain_policy', PolicyIdentifier),
+        ('subject_domain_policy', PolicyIdentifier),
     ]
 
 
@@ -1632,6 +1575,11 @@ class SubjectInfoAccessSyntax(SequenceOf):
     _child_spec = AccessDescription
 
 
+# https://tools.ietf.org/html/rfc7633
+class Features(SequenceOf):
+    _child_spec = Integer
+
+
 class EntrustVersionInfo(Sequence):
     _fields = [
         ('entrust_vers', GeneralString),
@@ -1672,6 +1620,8 @@ class ExtensionId(ObjectIdentifier):
         '2.5.29.54': 'inhibit_any_policy',
         '1.3.6.1.5.5.7.1.1': 'authority_information_access',
         '1.3.6.1.5.5.7.1.11': 'subject_information_access',
+        # https://tools.ietf.org/html/rfc7633
+        '1.3.6.1.5.5.7.1.24': 'tls_feature',
         '1.3.6.1.5.5.7.48.1.5': 'ocsp_no_check',
         '1.2.840.113533.7.65.0': 'entrust_version_extension',
         '2.16.840.1.113730.1.1': 'netscape_certificate_type',
@@ -1705,6 +1655,7 @@ class Extension(Sequence):
         'inhibit_any_policy': Integer,
         'authority_information_access': AuthorityInfoAccessSyntax,
         'subject_information_access': SubjectInfoAccessSyntax,
+        'tls_feature': Features,
         'ocsp_no_check': Null,
         'entrust_version_extension': EntrustVersionInfo,
         'netscape_certificate_type': NetscapeCertificateType,
@@ -1764,6 +1715,7 @@ class Certificate(Sequence):
     _extended_key_usage_value = None
     _authority_information_access_value = None
     _subject_information_access_value = None
+    _tls_feature_value = None
     _ocsp_no_check_value = None
     _issuer_serial = None
     _authority_issuer_serial = False
@@ -2052,6 +2004,20 @@ class Certificate(Sequence):
             self._set_extensions()
         return self._subject_information_access_value
 
+    @property
+    def tls_feature_value(self):
+        """
+        This extension is used to list the TLS features a server must respond
+        with if a client initiates a request supporting them.
+
+        :return:
+            None or a Features object
+        """
+
+        if not self._processed_extensions:
+            self._set_extensions()
+        return self._tls_feature_value
+
     @property
     def ocsp_no_check_value(self):
         """
@@ -2546,122 +2512,3 @@ class Certificate(Sequence):
             return True
 
         return False
-
-
-def _iri_utf8_errors_handler(exc):
-    """
-    Error handler for decoding UTF-8 parts of a URI into an IRI. Leaves byte
-    sequences encoded in %XX format, but as part of a unicode string.
-
-    :param exc:
-        The UnicodeDecodeError exception
-
-    :return:
-        A 2-element tuple of (replacement unicode string, integer index to
-        resume at)
-    """
-
-    bytes_as_ints = bytes_to_list(exc.object[exc.start:exc.end])
-    replacements = ['%%%02x' % num for num in bytes_as_ints]
-    return (''.join(replacements), exc.end)
-
-
-codecs.register_error('iriutf8', _iri_utf8_errors_handler)
-
-
-def _urlquote(string, safe=''):
-    """
-    Quotes a unicode string for use in a URL
-
-    :param string:
-        A unicode string
-
-    :param safe:
-        A unicode string of character to not encode
-
-    :return:
-        None (if string is None) or an ASCII byte string of the quoted string
-    """
-
-    if string is None or string == '':
-        return None
-
-    # Anything already hex quoted is pulled out of the URL and unquoted if
-    # possible
-    escapes = []
-    if re.search('%[0-9a-fA-F]{2}', string):
-        # Try to unquote any percent values, restoring them if they are not
-        # valid UTF-8. Also, requote any safe chars since encoded versions of
-        # those are functionally different than the unquoted ones.
-        def _try_unescape(match):
-            byte_string = unquote_to_bytes(match.group(0))
-            unicode_string = byte_string.decode('utf-8', 'iriutf8')
-            for safe_char in list(safe):
-                unicode_string = unicode_string.replace(safe_char, '%%%02x' % ord(safe_char))
-            return unicode_string
-        string = re.sub('(?:%[0-9a-fA-F]{2})+', _try_unescape, string)
-
-        # Once we have the minimal set of hex quoted values, removed them from
-        # the string so that they are not double quoted
-        def _extract_escape(match):
-            escapes.append(match.group(0).encode('ascii'))
-            return '\x00'
-        string = re.sub('%[0-9a-fA-F]{2}', _extract_escape, string)
-
-    output = urlquote(string.encode('utf-8'), safe=safe.encode('utf-8'))
-    if not isinstance(output, byte_cls):
-        output = output.encode('ascii')
-
-    # Restore the existing quoted values that we extracted
-    if len(escapes) > 0:
-        def _return_escape(_):
-            return escapes.pop(0)
-        output = re.sub(b'%00', _return_escape, output)
-
-    return output
-
-
-def _urlunquote(byte_string, remap=None, preserve=None):
-    """
-    Unquotes a URI portion from a byte string into unicode using UTF-8
-
-    :param byte_string:
-        A byte string of the data to unquote
-
-    :param remap:
-        A list of characters (as unicode) that should be re-mapped to a
-        %XX encoding. This is used when characters are not valid in part of a
-        URL.
-
-    :param preserve:
-        A bool - indicates that the chars to be remapped if they occur in
-        non-hex form, should be preserved. E.g. / for URL path.
-
-    :return:
-        A unicode string
-    """
-
-    if byte_string is None:
-        return byte_string
-
-    byte_string = unquote_to_bytes(byte_string)
-
-    if preserve:
-        replacements = ['\x1A', '\x1C', '\x1D', '\x1E', '\x1F']
-        preserve_unmap = {}
-        for char in remap:
-            replacement = replacements.pop(0)
-            preserve_unmap[replacement] = char
-            byte_string = byte_string.replace(char.encode('ascii'), replacement.encode('ascii'))
-
-    if remap:
-        for char in remap:
-            byte_string = byte_string.replace(char.encode('ascii'), ('%%%02x' % ord(char)).encode('ascii'))
-
-    output = byte_string.decode('utf-8', 'iriutf8')
-
-    if preserve:
-        for replacement, original in preserve_unmap.items():
-            output = output.replace(replacement, original)
-
-    return output
